/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/
/* This file contains the description of the API between the base radius_gw extension and
  its extensions that add support for specific RADIUS attributes and/or commands.
  The "notes.txt" file contains the basic mechanism for use of this API. */

/* Note that waaad/waaad.h must be included before this file */  
#ifndef _RADIUS_GW_H
#define _RADIUS_GW_H

#include <stdint.h>	/* uint8_t, etc... */
#include <waaad/waaad.h> /* session_t, etc... */

/* This type is used for all lists in this extension */
struct rg_list {
	struct rg_list *next;
	struct rg_list *prev;
	struct rg_list *head;
};

/**************************************************************/
/*                  RADIUS messages                           */
/**************************************************************/

/* Note on design: the parsing of RADIUS message is not very efficient since we have to duplicate the memory of all
 attributes, instead of pointing back to them in the original message. Anyway, it makes adding / removing attributes simpler. */

/* This type describes a RADIUS attribute */
struct rad_attr {
	/* Meta data */
	struct rg_list	chain;	/* link this attribute in a message */
	int		handled; /* Has this attribute already been converted to Diameter? */
	
	/* Data */
	uint8_t		type;
	uint8_t		length;
	union 	{
		uint8_t	buf[253];
		struct {
			uint32_t	vendor_id;
			union {
				uint8_t	string[249];			/* generic format */
				struct {
					uint8_t	vendor_type;
					uint8_t	vendor_length;
					uint8_t vendor_value[247];
				} 	tlv;				/* TLV format defined in rfc2865#section-5.26 */
				struct {
					unsigned	m 	:1;
					unsigned	tag	:7;
					uint8_t		data[248];
				} radext;				/* Extended attributes defined in draft-ietf-radext-extended-attributes-08 */
			};
		} 	vsa; /* vendor-specific attributes */
	}		data; /* Always fits in 253 bytes */
};
	

/* The following type represents a complete RADIUS message (internal representation) with parsing information */
typedef struct _rad_t {
	/* Metadata */
	struct rg_list	attributes;	/* The list of attributes */
		
	/* Data */
	uint8_t		code;
	uint8_t		identifier;
	uint8_t		authenticator[16];
} rad_t;


/**************************************************************/
/*                  Extensions registration                   */
/**************************************************************/

#define RADIUS_GW_API_VER	1 /* increment when making changes to radius_gw_api definition bellow */
struct radius_gw_api {
	void *  (*rga_conf_parse_cb) ( char * conf_file );	/* configuration parser. Returns NULL on error only. Called even if no configuration file is passed (with NULL parameter then) */
	void	(*rga_conf_free_cb) (void * conf); 		/* Free an object returned by previous cb */
	
	int	(*rga_rad_req_cb) ( void * conf, sess_id_t ** session, rad_t ** rad_req, msg_t ** diam_fw ); /* handle an incoming RADIUS message */
	int	(*rga_diam_ans_cb) ( void * conf, sess_id_t ** session, msg_t ** diam_ans, rad_t ** rad_fw ); /* handle the corresponding Diameter answer */
};

/* All extensions must provide the following entry point that is called when the extension is loaded.
Beware, the same extension may be loaded several times, and receive different configuration files. 
No global data should be initialized during this function; instead it should be done during the rga_conf_parse_cb call,
and store in the memory pointed by "conf" that is passed in turn to all callbacks. */
extern int rga_register(int version, waaad_api_t * waaad_api, struct radius_gw_api * api);


/**************************************************************/
/*      Functions exported by the common library extension    */
/**************************************************************/

/* List management */

void rg_list_init(struct rg_list * plist);
int  rg_list_is_empty(struct rg_list * plist);
void rg_list_insert_after(struct rg_list * ref, struct rg_list * item);
void rg_list_insert_before(struct rg_list * ref, struct rg_list * item);
void rg_list_unlink(struct rg_list * plist);


/* Radius message */
void rg_msg_free(rad_t * msg);


/****************************************/
/*      Debug and related stuff         */
/****************************************/

#include <assert.h>
#ifndef ASSERT
#define ASSERT(x) assert(x)
#endif /* ASSERT */


/* CHECK_* macro from the daemon */
/* Helper for tracing the CHECK_* macros bellow */
#define TRACE_DEBUG_ALL( str ) 	\
	TRACE_DEBUG(CALL, str );


/* Macros to check a return value and branch out in case of error.
 * These macro are to be used only when errors are highly improbable, not for expected errors.
 */

/* Check the return value of a system function and execute fallback in case of error */
#define CHECK_SYS_DO( __call__, __fallback__  ) { 					\
	int __ret__;									\
	TRACE_DEBUG_ALL( "Check SYS: " #__call__ );					\
	__ret__ = (__call__);								\
	if (__ret__ < 0) {								\
		int __err__ = errno;	/* We may handle EINTR here */			\
		log_error("An unexpected error occured (%s), turn on debug for detail\n",\
			strerror(__err__));						\
		TRACE_DEBUG(INFO, "Error in '" #__call__ "': %s", strerror(__err__));	\
		__fallback__;								\
	}										\
}
/* Check the return value of a system function, return error code on error */
#define CHECK_SYS( __call__  ) { 							\
	int __ret__;									\
	TRACE_DEBUG_ALL( "Check SYS: " #__call__ );					\
	__ret__ = (__call__);								\
	if (__ret__ < 0) {								\
		int __err__ = errno;	/* We may handle EINTR here */			\
		log_error("An unexpected error occured (%s), turn on debug for detail\n",\
			strerror(__err__));						\
		TRACE_DEBUG(INFO, "Error in '" #__call__ "': %s", strerror(__err__));	\
		return __err__;								\
	}										\
}

/* Check the return value of a POSIX function and execute fallback in case of error or special value */
#define CHECK_POSIX_DO2( __call__, __val__, __fallback1__, __fallback2__ ) {			\
	int __ret__;										\
	TRACE_DEBUG_ALL( "Check POSIX: " #__call__ );						\
	__ret__ = (__call__);									\
	if (__ret__ != 0) {									\
		if (__ret__ == (__val__)) {							\
			__fallback1__;								\
		} else {									\
			log_error("An unexpected error, see log for detail\n");			\
			TRACE_DEBUG(INFO, "Error in '" #__call__ "': %s", strerror(__ret__));	\
			__fallback2__;								\
		}										\
	}											\
}

/* Check the return value of a POSIX function and execute fallback in case of error */
#define CHECK_POSIX_DO( __call__, __fallback__ ) 					\
	CHECK_POSIX_DO2( (__call__), 0, , __fallback__ );

/* Check the return value of a POSIX function and return it if error */
#define CHECK_POSIX( __call__ ) { 							\
	int __v__;									\
	CHECK_POSIX_DO( __v__ = (__call__), return __v__ );				\
}

/* Check that a memory allocator did not return NULL, otherwise log an error and execute fallback */
#define CHECK_MALLOC_DO( __call__, __fallback__ ) { 					\
	void *  __ret__;								\
	TRACE_DEBUG_ALL( "Check MALLOC: " #__call__ );					\
	__ret__ = (void *)( __call__ );							\
	if (__ret__ == NULL) {								\
		int __err__ = errno;							\
		log_error("Memory allocation failed: %s\n", strerror(__err__));		\
		TRACE_DEBUG(INFO, "Error in '" #__call__ "': %s", strerror(__err__));	\
		__fallback__;								\
	}										\
}

/* Check that a memory allocator did not return NULL, otherwise return ENOMEM */
#define CHECK_MALLOC( __call__ )							\
	CHECK_MALLOC_DO( __call__, return ENOMEM );


/* The next functions can be used also for expected errors */

/* Check parameters at function entry, execute fallback on error */
#define CHECK_PARAMS_DO( __bool__, __fallback__ )					\
	TRACE_DEBUG_ALL( "Check PARAMS: " #__bool__ );					\
	if ( ! (__bool__) ) {								\
		TRACE_DEBUG(INFO, "Invalid parameter received in " #__bool__ );		\
		__fallback__;								\
	}
/* Check parameters at function entry, return EINVAL if the boolean is false (similar to assert) */
#define CHECK_PARAMS( __bool__ )							\
	CHECK_PARAMS_DO( __bool__, return EINVAL );

/* Check the return value of an internal function, log and propagate */
#define CHECK_FCT_DO( __call__, __fallback__ ) {					\
	int __ret__;									\
	TRACE_DEBUG_ALL( "Check FCT: " #__call__ );					\
	__ret__ = (__call__);								\
	if (__ret__ != 0) {								\
		TRACE_DEBUG(INFO, "Error in '" #__call__ "': %s", strerror(__ret__));	\
		__fallback__;								\
	}										\
}
/* Check the return value of a function call, return any error code */
#define CHECK_FCT( __call__ ) {								\
	int __v__;									\
	CHECK_FCT_DO( __v__ = (__call__), return __v__ );				\
}
		

#endif /* _RADIUS_GW_H */
  
