/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Manage the list of sub-extensions that provide handlers for RADIUS messages / attributes */

#include "radius_gw_internal.h"
#include <dlfcn.h>
#include <libgen.h>

/* List of extensions, in the order they are written in the configuration file. */
static struct rg_list ext_list;

/* Description of an extension entry */
struct ext_descr {
	struct rg_list		chain; 	/* chaining in the ext_list list */
	
	void * 			dlo;	/* object returned by dlopen of the extension, to use with dlclose later */
	struct radius_gw_api	api;	/* the callbacks registered by rga_register */
	void *			conf;	/* the parsed configuration returned by rga_conf_parse_cb */
	
	int			port;	/* this extension is called for messages received on this(these) port(s) only */
	unsigned char *		cc;	/* array of command codes, or NULL for any cc */
	size_t			cc_len; /* size of the previous array */
	
	char *			extname; /* baseame of the extension, for debug messages. To be freed when object is detroyed */
	char * 			conffile; /* configuration file passed to the extension, or "(null)". To be freed when object is destroyed */
};

/* Accelerators for each command code (one for each port). These accelerators are built on-demand, as a cache, after start_cache function has been called.  */
static struct rg_list ext_accel_auth, ext_accel_acct;

/* accelerator list, one per command code */
struct ext_accel {
	struct rg_list		chain;	/* link in the ext_accel_* list */
	unsigned char		ccode;	/* the command code of this accelerator. The previous list is ordered according to this value. We don't handle extended CC yet */
	struct rg_list		extensions; /* head for the list of extensions to be called for this command code. List of ext_accel_item items */
};

/* accelerator item, references to extensions */
struct ext_accel_item {
	struct rg_list		chain; 	/* link in the ext_accel "extensions" list */
	struct ext_descr *	ext;	/* pointer to the extension data */
};

/* Mutex to protect all the previous lists */
static pthread_mutex_t ext_mtx = PTHREAD_MUTEX_INITIALIZER;

/* Has start_cache been called? */
static int cache_started;


/* The lock must be held before calling this function */
static int get_accelerator(struct rg_list ** list, unsigned char ccode, int port)
{
	struct rg_list *refer, *search;
	struct ext_accel * accel = NULL;
	struct ext_accel_item * item = NULL;
	
	TRACE_ENTRY("%p %hhu %i", list, ccode, port);
	
	CHECK_PARAMS( cache_started && list && ((port == RGW_EXT_PORT_AUTH) || (port == RGW_EXT_PORT_ACCT)) );
	
	if (port == RGW_EXT_PORT_AUTH)
		refer = &ext_accel_auth;
	else
		refer = &ext_accel_acct;
	
	/* Check if we have already an accelerator for this ccode */
	for (search = refer->next; search != refer; search = search->next) {
		struct ext_accel * loc = (struct ext_accel *)search;
		
		if (loc->ccode < ccode)
			continue;
		
		if (loc->ccode > ccode)
			break; /* we don't have an accelerator for this value yet */
		
		/* We found the matching accelerator, just return this list */
		*list = &loc->extensions;
		return 0;
	}
	
	/* We must create the accelerator list, then save it just before search */
	
	CHECK_MALLOC( accel = malloc(sizeof(struct ext_accel)) );
	memset(accel, 0, sizeof(struct ext_accel) );
	rg_list_init(&accel->chain);
	rg_list_init(&accel->extensions);
	accel->ccode = ccode;
	
	/* Check if each extension from the global list is enabled for this port and ccode */
	for (refer = ext_list.next; refer != &ext_list; refer = refer->next) {
		struct ext_descr * loc = (struct ext_descr *)refer;
		
		/* Skip if this extension is not registered for this port */
		if (! (loc->port & port) )
			continue;
		
		/* Check if the ccode is there */
		if (loc->cc) {
			int i;
			int match = 0;
			for (i=0; i< loc->cc_len; i++) {
				if (loc->cc[i] < ccode)
					continue;
				if (loc->cc[i] == ccode)
					match = 1;
				break;
			}
			if (!match)
				continue;
		}
		
		/* Ok, this extension is called for this port / ccode, add to the accelerator */
		CHECK_MALLOC( item = malloc(sizeof(struct ext_accel_item)) );
		memset(item, 0, sizeof(struct ext_accel_item));
		rg_list_init(&item->chain);
		item->ext = loc;
		/* Add as last element of the accelerator */
		rg_list_insert_before(&accel->extensions, &item->chain);
	}
	
	/* Now, save this accelerator entry in the global list */
	rg_list_insert_before(search, &accel->chain);
	
	return 0;
}


int rgw_extensions_init(void)
{
	TRACE_ENTRY();
	
	cache_started = 0;
	rg_list_init(&ext_list);
	rg_list_init(&ext_accel_auth);
	rg_list_init(&ext_accel_acct);
	
	return 0;
}

int rgw_extensions_add( char * extfile, char * conffile, int port, unsigned char ** codes_array, size_t codes_sz )
{
	struct ext_descr * new;
	int (* ext_rga_register)(int version, waaad_api_t * waaad, struct radius_gw_api * api);
	int ret = 0;
	char * myextfile = NULL;
	
	TRACE_ENTRY();
	
	CHECK_PARAMS( extfile && port && codes_array && (cache_started == 0) );
	
	CHECK_MALLOC( myextfile = strdup(extfile) );
	
	CHECK_MALLOC( new = malloc(sizeof(struct ext_descr)) );
	memset(new, 0, sizeof(struct ext_descr));
	
	rg_list_init(&new->chain);
	
	/* Copy some names to be freed when object is destroyed */
	CHECK_MALLOC( new->extname = strdup(basename(myextfile)) );
	free(myextfile);
	CHECK_MALLOC( new->conffile = conffile ? conffile : strdup("(null)") );
	
	/* Try and load the extension */
	TRACE_DEBUG(INFO, "Loading subextension: %s", extfile);
	new->dlo = dlopen(extfile, RTLD_LAZY | RTLD_LOCAL);
	if (new->dlo == NULL) {
		/* An error occured */
		log_error("Loading of subextension %s failed:\n %s\n", extfile, dlerror());
		goto error;
	}
	
	/* Resolve the entry point */
	ext_rga_register = dlsym( new->dlo, "rga_register" );
	if (ext_rga_register == NULL) {
		/* An error occured */
		log_error("Unable to resolve 'rga_register' in subextension %s:\n %s\n", extfile, dlerror());
		goto error;
	}
	
	/* Call the entry point */
	TRACE_DEBUG(FULL, "Calling subextension entry point...");
	CHECK_FCT_DO(  (*ext_rga_register) (RADIUS_GW_API_VER, waaad_api, &new->api),  goto error  );
	
	/* Now parse the configuration file, this will initialize all extension states and store it in the returned pointer (the subextensions must be re-entrant) */
	TRACE_DEBUG(FULL, "Parsing subext conf file: %s", new->conffile );
	new->conf = (*(new->api.rga_conf_parse_cb))(conffile);
	if (new->conf == NULL) {
		log_error("An error occurred while parsing configuration parameter for extension '%s' (%s), aborting...\n", new->extname, new->conffile);
		goto error;
	}
	
	/* Now order the array of command codes and save */
	if (*codes_array && codes_sz) {
		int i;
		
		new->cc = *codes_array;
		*codes_array = NULL;
		
		for (i = 0; i < codes_sz - 1; i++) {
			int j, idx = i, min = new->cc[i];
			
			/* find the smallest remaining element */
			for (j = i + 1; j < codes_sz; j++) {
				if (min > new->cc[j]) {
					min = new->cc[j];
					idx = j;
				}
			}
			
			/* swap if needed */
			if (idx != i) {
				int tmp = new->cc[i];
				new->cc[i] = new->cc[idx];
				new->cc[idx] = tmp;
			}
		}
		new->cc_len = codes_sz;
	}
		
	new->port = port;
	
	/* And save this new extension in the list */
	CHECK_POSIX( pthread_mutex_lock(&ext_mtx) );
	rg_list_insert_before(&ext_list, &new->chain);
	CHECK_POSIX( pthread_mutex_unlock(&ext_mtx) );
	
	return 0;
	
	
error:
	if (new && new->dlo)
		dlclose(new->dlo);
	if (new)
		free(new);	
	return EINVAL;
}

void rgw_extensions_dump(void)
{
	struct ext_descr * ext;
	struct rg_list * ptr, *ptraccel;
	
	if ( ! TRACE_BOOL(FULL) )
		return;
	
	CHECK_POSIX_DO( pthread_mutex_lock(&ext_mtx), );
	if ( ! rg_list_is_empty( &ext_list ) )
		log_debug(" RADIUS gateway registered sub-extensions:\n");
	
	for (ptr = ext_list.next; ptr != &ext_list; ptr = ptr->next) {
		char * codes = NULL;
		
		ext = (struct ext_descr *)ptr;
		if (ext->cc) {
			int i;
			char * nxt;
			
			CHECK_MALLOC_DO( codes = malloc(ext->cc_len * 3 + 1), break );
			nxt = codes;
			for (i = 0; i < ext->cc_len; i++) {
				sprintf(nxt, "%02hhx ", ext->cc[i]);
				nxt += 3;
			}
		}
		
		log_debug("  %-15s (%-20s) - p: %s %s, %d cc: %s\n", 
				ext->extname, 
				ext->conffile, 
				ext->port & RGW_EXT_PORT_AUTH ? "Au" : "  ",
				ext->port & RGW_EXT_PORT_ACCT ? "Ac" : "  ",
				ext->cc_len,
				ext->cc ? codes : "*");
		
		free(codes);
	}
	
	CHECK_POSIX_DO( pthread_mutex_unlock(&ext_mtx), );
	
	
	/* Dump the list of accelerators */
	if ( ! TRACE_BOOL(FULL + 1) )
		return;
	
	CHECK_POSIX_DO( pthread_mutex_lock(&ext_mtx), );
	if ( !rg_list_is_empty( &ext_accel_auth ) || !rg_list_is_empty( &ext_accel_acct ))
		log_debug(" RADIUS gateway sub-extensions accelerators:\n");
	
	for (ptraccel = ext_accel_auth.next; ptraccel != &ext_accel_auth; ptraccel = ptraccel->next) {
		struct ext_accel * accel = (struct ext_accel *)ptraccel;
		log_debug("  auth, code %02hhu:\n", accel->ccode);

		for (ptr = accel->extensions.next; ptr != &accel->extensions; ptr = ptr->next) {
			struct ext_accel_item * item = (struct ext_accel_item *)ptr;
			log_debug("     %-15s (%s)\n", item->ext->extname, item->ext->conffile);
		}
	}
	for (ptraccel = ext_accel_acct.next; ptraccel != &ext_accel_acct; ptraccel = ptraccel->next) {
		struct ext_accel * accel = (struct ext_accel *)ptraccel;
		log_debug("  acct, code %02hhu:\n", accel->ccode);

		for (ptr = accel->extensions.next; ptr != &accel->extensions; ptr = ptr->next) {
			struct ext_accel_item * item = (struct ext_accel_item *)ptr;
			log_debug("     %-15s (%s)\n", item->ext->extname, item->ext->conffile);
		}
	}
	
	
	CHECK_POSIX_DO( pthread_mutex_unlock(&ext_mtx), );
	
}

void rgw_extensions_start_cache(void)
{
	cache_started++;
	
	/* tmp: for debug purpose
	{
		struct rg_list *mylist = NULL;
		CHECK_FCT_DO( get_accelerator(&mylist, 4, RGW_EXT_PORT_AUTH),  );
		CHECK_FCT_DO( get_accelerator(&mylist, 4, RGW_EXT_PORT_ACCT),  );
		CHECK_FCT_DO( get_accelerator(&mylist, 8, RGW_EXT_PORT_AUTH),  );
		CHECK_FCT_DO( get_accelerator(&mylist, 8, RGW_EXT_PORT_ACCT),  );
		CHECK_FCT_DO( get_accelerator(&mylist, 0x0b, RGW_EXT_PORT_AUTH),  );
		CHECK_FCT_DO( get_accelerator(&mylist, 0x0b, RGW_EXT_PORT_ACCT),  );
		rgw_extensions_dump();
	}
	*/
}

void rgw_extensions_fini(void)
{
	struct rg_list * item, *subitem;
	
	TRACE_ENTRY();
	
	CHECK_POSIX_DO( pthread_mutex_lock(&ext_mtx), );
	
	/* Remove all elements from all accelerators */
	while ( ! rg_list_is_empty(&ext_accel_auth) ) {
		item = ext_accel_auth.next;
		rg_list_unlink(item);
		{
			struct ext_accel * accel = (struct ext_accel *)item;
			while ( ! rg_list_is_empty(&accel->extensions) ) {
				subitem = accel->extensions.next;
				rg_list_unlink(subitem);
				free(subitem);
			}
		}
		free(item);
	}
	while ( ! rg_list_is_empty(&ext_accel_acct) ) {
		item = ext_accel_acct.next;
		rg_list_unlink(item);
		{
			struct ext_accel * accel = (struct ext_accel *)item;
			while ( ! rg_list_is_empty(&accel->extensions) ) {
				subitem = accel->extensions.next;
				rg_list_unlink(subitem);
				free(subitem);
			}
		}
		free(item);
	}
	
	/* Now destroy all extensions */
	while ( ! rg_list_is_empty(&ext_list) ) {
		struct ext_descr * ext = (struct ext_descr *) ext_list.next;
		rg_list_unlink(&ext->chain);
		free(ext->conffile);
		free(ext->extname);
		free(ext->cc);
		if (ext->conf)
			(*ext->api.rga_conf_free_cb)(ext->conf);
		dlclose(ext->dlo);
		free(ext);
	}
	
	CHECK_POSIX_DO( pthread_mutex_unlock(&ext_mtx), );
}
